@@ -1,3 +1,4 @@
+# frozen_string_literal: true
 # coding: ASCII-8BIT
 
 # FastImage finds the size or type of an image given its uri.
@@ -28,6 +29,9 @@
 # or referrer which some servers require. Pass an :http_header argument to specify headers,
 # e.g., :http_header => {'User-Agent' => 'Fake Browser'}.
 #
+# FastImage can give you information about the parsed display orientation of an image with Exif
+# data (jpeg or tiff).
+#
 # === Examples
 #   require 'fastimage'
 #
@@ -39,9 +43,12 @@
 #   => :gif
 #   File.open("/some/local/file.gif", "r") {|io| FastImage.type(io)}
 #   => :gif
+#   FastImage.new("http://stephensykes.com/images/pngimage").content_length
+#   => 432
+#   FastImage.new("http://stephensykes.com/images/ExifOrientation3.jpg").orientation
+#   => 3
 #
 # === References
-# * http://snippets.dzone.com/posts/show/805
 # * http://www.anttikupila.com/flash/getting-jpg-dimensions-with-as3-without-loading-the-entire-file/
 # * http://pennysmalls.wordpress.com/2008/08/19/find-jpeg-dimensions-fast-in-pure-ruby-no-ima/
 # * https://rubygems.org/gems/imagesize
@@ -49,14 +56,21 @@
 #
 
 require 'net/https'
-require 'addressable/uri'
-require 'fastimage/fbr.rb'
 require 'delegate'
 require 'pathname'
 require 'zlib'
+require 'base64'
+require 'uri'
+
+# see http://stackoverflow.com/questions/5208851/i/41048816#41048816
+if RUBY_VERSION < "2.2"
+  module URI
+    DEFAULT_PARSER = Parser.new(:HOSTNAME => "(?:(?:[a-zA-Z\\d](?:[-\\_a-zA-Z\\d]*[a-zA-Z\\d])?)\\.)*(?:[a-zA-Z](?:[-\\_a-zA-Z\\d]*[a-zA-Z\\d])?)\\.?")
+  end
+end
 
 class FastImage
-  attr_reader :size, :type, :content_length
+  attr_reader :size, :type, :content_length, :orientation
 
   attr_reader :bytes_read
 
@@ -177,12 +191,16 @@ class FastImage
 
     @property = @options[:type_only] ? :type : :size
 
+    @type, @state = nil
+
     if uri.respond_to?(:read)
       fetch_using_read(uri)
+    elsif uri.start_with?('data:')
+      fetch_using_base64(uri)
     else
       begin
-        @parsed_uri = Addressable::URI.parse(uri)
-      rescue Addressable::URI::InvalidURIError
+        @parsed_uri = URI.parse(uri)
+      rescue URI::InvalidURIError
         fetch_using_file_open
       else
         if @parsed_uri.scheme == "http" || @parsed_uri.scheme == "https"
@@ -193,12 +211,10 @@ class FastImage
       end
     end
 
-    uri.rewind if uri.respond_to?(:rewind)
-
     raise SizeNotFound if @options[:raise_on_failure] && @property == :size && !@size
 
   rescue Timeout::Error, SocketError, Errno::ECONNREFUSED, Errno::EHOSTUNREACH, Errno::ECONNRESET,
-    ImageFetchFailure, Net::HTTPBadResponse, EOFError, Errno::ENOENT
+    ImageFetchFailure, Net::HTTPBadResponse, EOFError, Errno::ENOENT, OpenSSL::SSL::SSLError
     raise ImageFetchFailure if @options[:raise_on_failure]
   rescue NoMethodError  # 1.8.7p248 can raise this due to a net/http bug
     raise ImageFetchFailure if @options[:raise_on_failure]
@@ -213,6 +229,9 @@ class FastImage
       end
     end
 
+  ensure
+    uri.rewind if uri.respond_to?(:rewind)
+
   end
 
   private
@@ -223,6 +242,17 @@ class FastImage
     fetch_using_http_from_parsed_uri
   end
 
+  # Some invalid locations need escaping
+  def escaped_location(location)
+    begin
+      URI(location)
+    rescue URI::InvalidURIError
+      URI.escape(location)
+    else
+      location
+    end
+  end
+
   def fetch_using_http_from_parsed_uri
     http_header = {'Accept-Encoding' => 'identity'}.merge(@options[:http_header])
 
@@ -231,14 +261,8 @@ class FastImage
       if res.is_a?(Net::HTTPRedirection) && @redirect_count < 4
         @redirect_count += 1
         begin
-          newly_parsed_uri = Addressable::URI.parse(res['Location'])
-          # The new location may be relative - check for that
-          if newly_parsed_uri.scheme != "http" && newly_parsed_uri.scheme != "https"
-            @parsed_uri.path = res['Location']
-          else
-            @parsed_uri = newly_parsed_uri
-          end
-        rescue Addressable::URI::InvalidURIError
+          @parsed_uri = URI.join(@parsed_uri, escaped_location(res['Location']))
+        rescue URI::InvalidURIError
         else
           fetch_using_http_from_parsed_uri
           break
@@ -276,14 +300,18 @@ class FastImage
     end
   end
 
+  def protocol_relative_url?(url)
+    url.start_with?("//")
+  end
+
   def proxy_uri
     begin
       if @options[:proxy]
-        proxy = Addressable::URI.parse(@options[:proxy])
+        proxy = URI.parse(@options[:proxy])
       else
-        proxy = ENV['http_proxy'] && ENV['http_proxy'] != "" ? Addressable::URI.parse(ENV['http_proxy']) : nil
+        proxy = ENV['http_proxy'] && ENV['http_proxy'] != "" ? URI.parse(ENV['http_proxy']) : nil
       end
-    rescue Addressable::URI::InvalidURIError
+    rescue URI::InvalidURIError
       proxy = nil
     end
     proxy
@@ -293,9 +321,9 @@ class FastImage
     proxy = proxy_uri
 
     if proxy
-      @http = Net::HTTP::Proxy(proxy.host, proxy.port).new(@parsed_uri.host, @parsed_uri.inferred_port)
+      @http = Net::HTTP::Proxy(proxy.host, proxy.port).new(@parsed_uri.host, @parsed_uri.port)
     else
-      @http = Net::HTTP.new(@parsed_uri.host, @parsed_uri.inferred_port)
+      @http = Net::HTTP.new(@parsed_uri.host, @parsed_uri.port)
     end
     @http.use_ssl = (@parsed_uri.scheme == "https")
     @http.verify_mode = OpenSSL::SSL::VERIFY_NONE
@@ -328,6 +356,7 @@ class FastImage
   end
 
   def fetch_using_file_open
+    @content_length = File.size?(@uri)
     File.open(@uri) do |s|
       fetch_using_read(s)
     end
@@ -339,6 +368,13 @@ class FastImage
     begin
       result = send("parse_#{@property}")
       if result
+        # extract exif orientation if it was found
+        if @property == :size && result.size == 3
+          @orientation = result.pop
+        else
+          @orientation = 1
+        end
+
         instance_variable_set("@#{@property}", result)
       else
         raise CannotParseImage
@@ -353,6 +389,13 @@ class FastImage
     send("parse_size_for_#{@type}")
   end
 
+  def fetch_using_base64(uri)
+    data = uri.split(',')[1]
+    decoded = Base64.decode64(data)
+    @content_length = decoded.size
+    fetch_using_read StringIO.new(decoded)
+  end
+
   module StreamUtil # :nodoc:
     def read_byte
       read(1)[0].ord
@@ -382,6 +425,7 @@ class FastImage
       @str = ''
     end
 
+    # Peeking beyond the end of the input will raise
     def peek(n)
       while @strpos + n - 1 >= @str.size
         unused_str = @str[@strpos..-1]
@@ -395,7 +439,7 @@ class FastImage
         @strpos = 0
       end
 
-      result = @str[@strpos..(@strpos + n - 1)]
+      @str[@strpos..(@strpos + n - 1)]
     end
 
     def read(n)
@@ -404,6 +448,24 @@ class FastImage
       @pos += n
       result
     end
+
+    def skip(n)
+      discarded = 0
+      fetched = @str[@strpos..-1].size
+      while n > fetched
+        discarded += @str[@strpos..-1].size
+        new_string = @read_fiber.resume
+        raise CannotParseImage if !new_string
+
+        new_string.force_encoding("ASCII-8BIT") if String.method_defined? :force_encoding
+
+        fetched += new_string.size
+        @str = new_string
+        @strpos = 0
+      end
+      @strpos = @strpos + n - discarded
+      @pos += n
+    end
   end
 
   class IOStream < SimpleDelegator # :nodoc:
@@ -411,7 +473,7 @@ class FastImage
   end
 
   def parse_type
-    case @stream.peek(2)
+    parsed_type = case @stream.peek(2)
     when "BM"
       :bmp
     when "GI"
@@ -421,7 +483,12 @@ class FastImage
     when 0x89.chr + "P"
       :png
     when "II", "MM"
-      :tiff
+      case @stream.peek(11)[8..10]
+      when "APC", "CR\002"
+        nil  # do not recognise CRW or CR2 as tiff
+      else
+        :tiff
+      end
     when '8B'
       :psd
     when "\0\0"
@@ -431,26 +498,27 @@ class FastImage
       when 2 then :cur
       end
     when "RI"
-      if @stream.peek(12)[8..11] == "WEBP"
-        :webp
-      else
-        raise UnknownImageType
-      end
+      :webp if @stream.peek(12)[8..11] == "WEBP"
     when "<s"
       :svg
-    when "<?"
-      if @stream.peek(100).include?("<svg")
-        :svg
-      else
-        raise UnknownImageType
+    when /<[?!]/
+      # Peek 10 more chars each time, and if end of file is reached just raise
+      # unknown. We assume the <svg tag cannot be within 10 chars of the end of
+      # the file, and is within the first 250 chars.
+      begin
+        :svg if (1..25).detect {|n| @stream.peek(10 * n).include?("<svg")}
+      rescue FiberError
+        nil
       end
-    else
-      raise UnknownImageType
     end
+
+    parsed_type or raise UnknownImageType
   end
 
   def parse_size_for_ico
-    @stream.read(8)[6..7].unpack('CC').map{|byte| byte == 0 ? 256 : byte }
+    icons = @stream.read(6)[4..5].unpack('v').first
+    sizes = icons.times.map { @stream.read(16).unpack('C2').map { |x| x == 0 ? 256 : x } }.sort_by { |w,h| w * h }
+    sizes.last
   end
   alias_method :parse_size_for_cur, :parse_size_for_ico
 
@@ -463,10 +531,11 @@ class FastImage
   end
 
   def parse_size_for_jpeg
+    exif = nil
     loop do
       @state = case @state
       when nil
-        @stream.read(2)
+        @stream.skip(2)
         :started
       when :started
         @stream.read_byte == 0xFF ? :sof : :started
@@ -478,7 +547,8 @@ class FastImage
           io = StringIO.new(data)
           if io.read(4) == "Exif"
             io.read(2)
-            @exif = Exif.new(IOStream.new(io)) rescue nil
+            new_exif = Exif.new(IOStream.new(io)) rescue nil
+            exif ||= new_exif # only use the first APP1 segment
           end
           :started
         when 0xe0..0xef
@@ -492,14 +562,14 @@ class FastImage
         end
       when :skipframe
         skip_chars = @stream.read_int - 2
-        @stream.read(skip_chars)
+        @stream.skip(skip_chars)
         :started
       when :readsize
-        s = @stream.read(3)
+        @stream.skip(3)
         height = @stream.read_int
         width = @stream.read_int
-        width, height = height, width if @exif && @exif.rotated?
-        return [width, height]
+        width, height = height, width if exif && exif.rotated?
+        return [width, height, exif ? exif.orientation : 1]
       end
     end
   end
@@ -520,7 +590,7 @@ class FastImage
 
   def parse_size_for_webp
     vp8 = @stream.read(16)[12..15]
-    len = @stream.read(4).unpack("V")
+    _len = @stream.read(4).unpack("V")
     case vp8
     when "VP8 "
       parse_size_vp8
@@ -539,7 +609,7 @@ class FastImage
   end
 
   def parse_size_vp8l
-    @stream.read(1) # 0x2f
+    @stream.skip(1) # 0x2f
     b1, b2, b3, b4 = @stream.read(4).bytes.to_a
     [1 + (((b2 & 0x3f) << 8) | b1), 1 + (((b4 & 0xF) << 10) | (b3 << 2) | ((b2 & 0xC0) >> 6))]
   end
@@ -558,14 +628,16 @@ class FastImage
   end
 
   class Exif # :nodoc:
-    attr_reader :width, :height
+    attr_reader :width, :height, :orientation
+
     def initialize(stream)
       @stream = stream
+      @width, @height, @orientation = nil
       parse_exif
     end
 
     def rotated?
-      @orientation && @orientation >= 5
+      @orientation >= 5
     end
 
     private
@@ -611,9 +683,15 @@ class FastImage
       @stream.read(2) # 42
 
       offset = @stream.read(4).unpack(@long)[0]
-      @stream.read(offset - 8)
+      if @stream.respond_to?(:skip)
+        @stream.skip(offset - 8)
+      else
+        @stream.read(offset - 8)
+      end
 
       parse_exif_ifd
+
+      @orientation ||= 1
     end
 
   end
@@ -621,9 +699,9 @@ class FastImage
   def parse_size_for_tiff
     exif = Exif.new(@stream)
     if exif.rotated?
-      [exif.height, exif.width]
+      [exif.height, exif.width, exif.orientation]
     else
-      [exif.width, exif.height]
+      [exif.width, exif.height, exif.orientation]
     end
   end
 
@@ -634,6 +712,7 @@ class FastImage
   class Svg # :nodoc:
     def initialize(stream)
       @stream = stream
+      @width, @height, @ratio, @viewbox_width, @viewbox_height = nil
       parse_svg
     end
 
@@ -644,6 +723,10 @@ class FastImage
         [@width, @width / @ratio]
       elsif @height && @ratio
         [@height * @ratio, @height]
+      elsif @viewbox_width && @viewbox_height
+        [@viewbox_width, @viewbox_height]
+      else
+        nil
       end
     end
 
@@ -666,14 +749,20 @@ class FastImage
             return if @width
           elsif attr_name.join =~ /viewbox/i
             values = attr_value.split(/\s/)
-            @ratio = values[2].to_f / values[3].to_f
+            if values[2].to_f > 0 && values[3].to_f > 0
+              @ratio = values[2].to_f / values[3].to_f
+              @viewbox_width = values[2].to_i
+              @viewbox_height = values[3].to_i
+            end
           end
         when /\w/
           attr_name << char
+        when "<"
+          attr_name = [char]
         when ">"
           state = :stop if state == :started
         else
-          state = :started if attr_name.join == "svg"
+          state = :started if attr_name.join == "<svg"
           attr_name.clear
         end
       end
